""" ADFS Checkpoint """

from xcputils.checkpoints import Checkpoint
from xcputils.ingestion.az import AdfsConnectionSettings, AdfsIngestor
from xcputils.ingestion.string import StringIngestor
from xcputils.streaming.az import AdfsStreamWriter


class AdfsCheckpoint(Checkpoint):
    """ ADFS Checkpoint """

    def __init__(
        self,
        name: str,
        container: str,
        directory: str,
        storage_account_name: str = None,
        tenant_id: str = None,
        client_id: str = None,
        client_secret: str = None,
        ):
        """ Constructor """
        self.adfs_connection_settings = AdfsConnectionSettings(
            container=container,
            file_name=f"{name}.json",
            directory=directory,
            storage_account_name=storage_account_name,
            tenant_id=tenant_id,
            client_id=client_id,
            client_secret=client_secret,
            )

    def get(self, default: str = None):
        """ Get checkpoint """
        try:
            checkpoint = AdfsIngestor(self.adfs_connection_settings).write_to_string()
        except:
            checkpoint = ""
        
        if checkpoint == "":
            checkpoint = default

        return checkpoint
    

    def set(self, value: str):
        """ Set checkpoint """
        StringIngestor(value, AdfsStreamWriter(self.adfs_connection_settings)).ingest()



""" Unit tests """

from datetime import datetime
import os
import unittest
from xcputils.checkpoints.file import FileCheckpoint


class TestFileCheckpoint(unittest.TestCase):
    """ Test xcputils.checkpoints.az.FileCheckpoint """


    def test_file_checkpoint(self):
        """ xcputils.checkpoints.az.FileCheckpoint """

        if not os.path.isdir("data"):
            os.mkdir("data")

        checkpoint = FileCheckpoint(
            name="dev.test.test_file_checkpoint",
            directory="data",
            )

        checkpoint.reset()
        expected = "default"
        actual = checkpoint.get(default=expected)
        self.assertEqual(actual, expected)

        expected = str(datetime.now())
        checkpoint.set(expected)
        actual = checkpoint.get()
        self.assertEqual(actual, expected)


if __name__ == '__main__':
    unittest.main()



""" ADFS Checkpoint """

from xcputils.checkpoints import Checkpoint
from xcputils.ingestion.aws import AwsS3Ingestor
from xcputils.ingestion.string import StringIngestor
from xcputils.streaming.aws import AwsS3ConnectionSettings, AwsS3StreamWriter


class AwsS3Checkpoint(Checkpoint):
    """ ADFS Checkpoint """

    def __init__(
        self,
        name: str,
        bucket: str,
        directory: str,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
        aws_session_token: str = None,
        aws_region_name: str = None,
        ):
        """ Constructor """
        self.awss3_connection_settings = AwsS3ConnectionSettings(
            bucket=bucket,
            file_path=f"{directory}/{name}.json",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            aws_region_name=aws_region_name,
            )

    def get(self, default: str = None):
        """ Get checkpoint """
        try:
            checkpoint = AwsS3Ingestor(self.awss3_connection_settings).write_to_string()
        except:
            checkpoint = ""
        
        if checkpoint == "":
            checkpoint = default

        return checkpoint

    def set(self, value: str):
        """ Set checkpoint """
        StringIngestor(value, AwsS3StreamWriter(self.awss3_connection_settings)).ingest()



""" String stream writer """

from xcputils.streaming import StreamWriter


class StringStreamWriter(StreamWriter):
    """ String stream writer """


    def __init__(self):
        super().__init__()
        self.value = ""


    def write(self, input_stream):
        """ Write to stream """

        content = input_stream if isinstance(input_stream, bytes) else  input_stream.read()
        if isinstance(content, bytes):
            content = content.decode('utf-8')
        if self.value:
            self.value += "\n"
        self.value += content



""" File ingestion """

from ftplib import FTP
from io import BytesIO
from xcputils.ingestion import Ingestor
from xcputils.streaming import StreamWriter


class FtpIngestor(Ingestor):
    """ Ingest from FTP """

    def __init__(
        self,
        url: str,
        folder: str = None,
        file_name: str = None,
        user: str = None,
        password: str = None,
        stream_writer: StreamWriter = None,
    ):

        super().__init__(stream_writer)

        self.url = url
        self.folder = folder
        self.file_name = file_name
        self.user = user
        self.password = password


    def __is_dir(self, ftp, filename):
        try:
            ftp.size(filename)
            return False
        except:
            return True


    def ingest(self):
        """ Ingest """

        ftp = FTP(host=self.url, user=self.user, passwd=self.password)

        if self.folder:
            ftp.cwd(self.folder)

        for file_name in ftp.nlst():
            if not self.__is_dir(ftp, file_name) \
                and (not self.file_name or self.file_name.lower() == file_name.lower()):
                hold_filename = self.stream_writer.get_file_path()
                self.stream_writer.set_file_path(
                    file_name if not hold_filename else f"{hold_filename}/{file_name}")
                ftp.retrbinary(f"RETR {file_name}", self.__write)
                self.stream_writer.set_file_path(hold_filename)

        ftp.quit()

    def __write(self, bytes):
        with BytesIO() as stream:
            stream.write(bytes)
            stream.seek(0)
            self.stream_writer.write(stream)



""" Ingestors """


from xcputils.streaming import StreamWriter
from xcputils.streaming.aws import AwsS3ConnectionSettings, AwsS3StreamWriter
from xcputils.streaming.az import AdfsConnectionSettings, AdfsStreamWriter
from xcputils.streaming.file import FileStreamWriter
from xcputils.streaming.string import StringStreamWriter


class Ingestor():
    """ Ingstor base class"""

    def __init__(self, stream_writer: StreamWriter = None):
        """ Constructor """

        self.stream_writer = stream_writer


    def write_to_aws_s3(
        self,
        bucket: str,
        file_path: str,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
        aws_session_token: str = None,
        aws_region_name: str = None):
        """ Write to AWS S3 """

        aws_s3_connection_settings = AwsS3ConnectionSettings(
            bucket=bucket,
            file_path=file_path,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            aws_region_name=aws_region_name)

        self.stream_writer = AwsS3StreamWriter(aws_s3_connection_settings)

        self.ingest()


    def write_to_adfs(
        self,
        container: str,
        file_name: str,
        directory: str,
        storage_account_name: str = None,
        tenant_id: str = None,
        client_id: str = None,
        client_secret: str = None,
        ):
        """ Write to Azure Data Lake Storage """
        adfs_connection_settings = AdfsConnectionSettings(
            container=container,
            file_name=file_name,
            directory=directory,
            storage_account_name=storage_account_name,
            tenant_id=tenant_id,
            client_id=client_id,
            client_secret=client_secret,
            )

        self.stream_writer = AdfsStreamWriter(adfs_connection_settings)

        self.ingest()


    def write_to_string(
        self,
        ) -> str:
        """ Write to Azure Data Lake Storage """

        self.stream_writer = StringStreamWriter()

        self.ingest()

        return self.stream_writer.value


    def write_to_file(self, file_path: str):
        """ Write to file """

        self.stream_writer = FileStreamWriter(file_path=file_path)

        self.ingest()


    def ingest(self):
        """ Ingest """



""" File ingestion """

from xcputils.ingestion import Ingestor
from xcputils.streaming import StreamWriter


class FileIngestor(Ingestor):
    """ Ingest from file """

    def __init__(
        self,
        file_path: str,
        stream_writer: StreamWriter = None,
        ):

        super().__init__(stream_writer)

        self.file_path = file_path


    def ingest(self):
        """ Ingest """

        with open(self.file_path, "rb") as file_stream:
            self.stream_writer.write(file_stream)



""" Unit tests """

from datetime import datetime
import os
import unittest
from xcputils.checkpoints.aws import AwsS3Checkpoint


class TestAwsS3Checkpoint(unittest.TestCase):
    """ Test xcputils.checkpoints.az.AwsS3Checkpoint """


    def test_adfs_checkpoint(self):
        """ xcputils.checkpoints.az.AwsS3Checkpoint """

        checkpoint = AwsS3Checkpoint(
            name="dev.test.test_awss3_checkpoint",
            bucket=os.environ['AWS_S3_BUCKET'],
            directory="testawss3checkpoint",
            )

        checkpoint.reset()
        expected = "default"
        actual = checkpoint.get(default=expected)
        self.assertEqual(actual, expected)

        expected = str(datetime.now())
        checkpoint.set(expected)
        actual = checkpoint.get()
        self.assertEqual(actual, expected)


if __name__ == '__main__':
    unittest.main()



""" Unit tests """

import datetime
import unittest
from xcputils.ingestion.az import AdfsIngestor
from xcputils.ingestion.string import StringIngestor
from xcputils.streaming.az import AdfsConnectionSettings


class TestAdfsIngestor(unittest.TestCase):
    """ Test xcputils.ingestion.az.AdfsIngestor """


    def test_ingest(self):
        """ xcputils.streaming.az.AdfsIngestor """

        container="testxcputils"
        directory="testadfsingestor"
        file_name="test.txt"
        payload = f"Testing.\n123.\næøåÆØÅ\n{datetime.datetime.now()}"


        StringIngestor(data=payload) \
            .write_to_adfs(
                container=container,
                directory=directory,
                file_name=file_name,
                )

        result = AdfsIngestor(
            AdfsConnectionSettings(
                container=container,
                directory=directory,
                file_name=file_name,
                )).write_to_string()

        self.assertEqual(result, payload)


if __name__ == '__main__':
    unittest.main()



""" Unit tests """

import json
import unittest
from unittest.mock import patch
from test.unit import mock_response
from requests.auth import HTTPBasicAuth
from requests import Session

from xcputils.ingestion.http import HttpIngestor, HttpMethod, HttpRequest


class TestHttpIngestor(unittest.TestCase):
    """ Test xcputils.ingest.http """


    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)


    @patch.object(Session, 'get')
    def test_get(self, mock_get):
        """ Test xcputils.ingest.get """

        mock_get.return_value = mock_response(json_data={"ip": "1.2.3.4"})
        result = HttpIngestor(
            http_request=HttpRequest(url="https://mock.com/ip")) \
            .write_to_string()
        self.assertTrue("ip" in result, f"key 'ip' not in: {result}")


    @patch.object(Session, 'get')
    def test_get_html(self, mock_get):
        """ Test xcputils.ingest.get HTML """

        mock_get.return_value = mock_response(content="<!DOCTYPE html>")
        result = HttpIngestor(
            http_request=HttpRequest(url="https://mock.com")) \
            .write_to_string()
        self.assertEqual(result[0:15], "<!DOCTYPE html>")


    @patch.object(Session, 'get')
    def test_auth(self, mock_get):
        """ Test xcputils.ingest.get authentication """

        mock_get.return_value = mock_response(json_data={"authenticated": True})
        result = HttpIngestor(
            http_request=HttpRequest(
                url="https://mock.com/basic-auth",
                auth=HTTPBasicAuth('postman', 'password'))) \
            .write_to_string()
        result = json.loads(result)
        self.assertEqual(result["authenticated"], True, result)


    @patch.object(Session, 'post')
    def test_post(self, mock_post):
        """ Test xcputils.ingest.post """

        data = {"test_key": "test_value"}

        mock_post.return_value = mock_response(json_data={"data" : data})
        result = HttpIngestor(
            http_request=HttpRequest(
                url="https://mock.com/post",
                method=HttpMethod.POST,
                body=data)
            ) \
            .write_to_string()
        result = json.loads(result)
        self.assertEqual(
            result["data"],
            data,
            f"Expected to contain data: {data}: {result}")


if __name__ == '__main__':
    unittest.main()



""" Unit tests """

from datetime import datetime
import unittest
from xcputils.checkpoints.az import AdfsCheckpoint


class TestAdfsCheckpoint(unittest.TestCase):
    """ Test xcputils.checkpoints.az.AdfsCheckpoint """


    def test_adfs_checkpoint(self):
        """ xcputils.checkpoints.az.AdfsCheckpoint """

        checkpoint = AdfsCheckpoint(
            name="dev.test.test_adfs_checkpoint",
            container="testxcputils",
            directory="testadfscheckpoint",
            )

        checkpoint.reset()
        expected = "default"
        actual = checkpoint.get(default=expected)
        self.assertEqual(actual, expected)

        expected = str(datetime.now())
        checkpoint.set(expected)
        actual = checkpoint.get()
        self.assertEqual(actual, expected)


if __name__ == '__main__':
    unittest.main()



""" File Checkpoint """

from xcputils.checkpoints import Checkpoint
from xcputils.ingestion.file import FileIngestor
from xcputils.ingestion.string import StringIngestor
from xcputils.streaming.file import FileStreamWriter


class FileCheckpoint(Checkpoint):
    """ File Checkpoint """

    def __init__(
        self,
        name: str,
        directory: str,
        ):
        """ Constructor """
        self.file_path=f"{directory}/{name}.json"

    def get(self, default: str = None) -> str:
        """ Get checkpoint """
        try:
            checkpoint = FileIngestor(self.file_path).write_to_string()
        except:
            checkpoint = ""
        
        if checkpoint == "":
            checkpoint = default

        return checkpoint

    def set(self, value: str):
        """ Set checkpoint """
        StringIngestor(value, FileStreamWriter(self.file_path)).ingest()



""" File stream writer """

from xcputils.streaming import StreamWriter


class FileStreamWriter(StreamWriter):
    """ File stream writer """

    def __init__(self, file_path: str):
        super().__init__()
        self.file_path = file_path


    def write(self, input_stream):
        """ Write stream to file """

        with open(file=self.file_path, mode="wb") as file_stream:
            file_stream.write(input_stream.read())

    def get_file_path(self) -> str:
        """ Get filename """

        return self.file_path


    def set_file_path(self, file_path: str):
        """ Set filename """

        self.file_path = file_path



""" Class Batch for async runs """

from __future__ import annotations
import asyncio
from typing import Coroutine, Callable
from functools import wraps, partial
import nest_asyncio

def async_wrap(func):
    """ Converts callable to async """
    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        pfunc = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, pfunc)
    return run


class Batch(list[Callable]):
    """ Runs tasks in batches asynchronously """


    def append(self, task: Callable, *args, **kwargs) -> Batch:
        """ Append task """
        atask = async_wrap(task)
        super().append(atask(*args, **kwargs))
        return self


    def append_async(self, task: Coroutine) -> Batch:
        """ Append async task """
        super().append(task)
        return self


    async def __execute_in_parrallel(self) -> Coroutine:
        tasks = []
        for coroutine in self:
            tasks.append(coroutine)
        return await asyncio.gather(*tasks, return_exceptions=True)        


    def __call__(self):
        nest_asyncio.apply()
        asyncio.set_event_loop(asyncio.new_event_loop())
        return asyncio.get_event_loop().run_until_complete(self.__execute_in_parrallel())


    def __repr__(self):
        return f"{self.__class__.__name__}({list.__repr__(self)})" 



""" Unit tests """

import json
import os
import unittest
from xcputils.ingestion.http import HttpIngestor, HttpRequest
from xcputils.streaming.aws import AwsS3ConnectionSettings, AwsS3StreamReader


class TestPaginatedHttpIngestor(unittest.TestCase):
    """ Test xcputils.ingest.http """


    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)


    def test_get(self):
        """ Test xcputils.ingest.http.get """

        HttpIngestor(
            http_request=HttpRequest(
                url="https://api.energidataservice.dk/dataset/CO2Emis",
                params={"start": "2022-01-01T00:00", "end": "2022-01-02T00:00"})
            ) \
            .with_pagination(
                page_size=100,
                data_property="records",
            ) \
            .write_to_aws_s3(
                bucket=os.environ['AWS_S3_BUCKET'],
                file_path="testpaginatedhttpingestor/eds/co2emis/co2emis.json"
                )

        stream_reader = AwsS3StreamReader(
            AwsS3ConnectionSettings(
                bucket=os.environ['AWS_S3_BUCKET'],
                file_path="testpaginatedhttpingestor/eds/co2emis/co2emis.1.json"
                )
            )

        page_1 = json.loads(stream_reader.read_str())
        self.assertTrue(len(page_1["records"]) == 100)

        stream_reader.connection_settings.file_path = \
            "testpaginatedhttpingestor/eds/co2emis/co2emis.2.json"
        page_6 = json.loads(stream_reader.read_str())
        self.assertTrue(len(page_6["records"]) == 100)


if __name__ == "__main__":
    unittest.main()



""" Unit tests """

import datetime
from io import BytesIO
import os
import unittest

from xcputils.streaming.aws import AwsS3StreamReader, AwsS3StreamWriter, AwsS3ConnectionSettings


class TestS3StreamWriter(unittest.TestCase):
    """ Test xcputils.streaming.aws.AwsS3StreamWriter """


    def test_write_read(self):
        """ xcputils.streaming.aws.S3StreamConnector write and read """

        connection_settings = AwsS3ConnectionSettings(
            bucket=os.environ['AWS_S3_BUCKET'],
            file_path="tests3streamwriter/test.txt")

        writer = AwsS3StreamWriter(connection_settings)

        payload = f"Testing.\n123.\næøåÆØÅ\n{datetime.datetime.now()}"

        with BytesIO() as stream:
            stream.write(payload.encode('utf-8'))
            stream.seek(0)
            writer.write(stream)

        reader = AwsS3StreamReader(connection_settings)

        actual_payload = reader.read_str()

        self.assertEqual(actual_payload, payload)


if __name__ == '__main__':
    unittest.main()



""" Azure Data Lake Storage Account Streaming Connector """

from typing import Any
import os
from azure.identity import DefaultAzureCredential, ClientSecretCredential
from azure.storage.filedatalake import DataLakeServiceClient

from xcputils.streaming import StreamReader, StreamWriter


class AdfsConnectionSettings():
    """ Azure Sata Lake Storage connection settings """

    def __init__(
        self,
        container: str,
        file_name: str,
        directory: str,
        storage_account_name: str = None,
        tenant_id: str = None,
        client_id: str = None,
        client_secret: str = None,
        ):

        self.container = container
        self.file_name = file_name
        self.directory = directory
        self.tenant_id = tenant_id
        self.client_id = client_id
        self.client_secret = client_secret

        if storage_account_name is None:
            storage_account_name = os.getenv("ADFS_DEFAULT_STORAGE_ACCOUNT", None)
        self.storage_account_name = storage_account_name


    def get_client(self):
        """ get ADFS client"""

        if self.tenant_id and self.client_id and self.client_secret:
            credential = ClientSecretCredential(
                tenant_id=self.tenant_id,
                client_id=self.client_id,
                client_secret=self.client_secret)
        else:
            credential = DefaultAzureCredential()

        return DataLakeServiceClient(
            account_url=f"https://{self.storage_account_name}.dfs.core.windows.net",
            credential=credential)


class AdfsStreamReader(StreamReader):
    """ Azure data Lake Storage stream reader """

    def __init__(self,
                 connection_settings: AdfsConnectionSettings):

        super().__init__()

        self.connection_settings = connection_settings


    def read(self, output_stream):
        """ Read from stream """

        client = self.connection_settings.get_client()

        file_system_client = client.get_file_system_client(
            file_system=self.connection_settings.container)

        directory_client = file_system_client.get_directory_client(
            self.connection_settings.directory)

        file_client = directory_client.get_file_client(self.connection_settings.file_name)

        downloader = file_client.download_file()

        downloader.readinto(output_stream)


class AdfsStreamWriter(StreamWriter):
    """ Azure data Lake Storage stream writer """

    def __init__(self,
                 connection_settings: AdfsConnectionSettings):

        super().__init__()

        self.connection_settings = connection_settings


    def get_file_path(self) -> str:
        """ Get filename """

        return self.connection_settings.file_name if not self.connection_settings.directory \
            else f"{self.connection_settings.directory}/{self.connection_settings.file_name}"


    def set_file_path(self, file_path: str):
        """ Set filename """

        directory, file_name = os.path.split(file_path)

        self.connection_settings.directory = directory
        self.connection_settings.file_name = file_name


    def write(self, input_stream: Any):
        """ Write to stream """

        client = self.connection_settings.get_client()
        file_system_client = client.get_file_system_client(
            file_system=self.connection_settings.container)

        if not file_system_client.exists():
            file_system_client = client.create_file_system(
                file_system=self.connection_settings.container)

        directory_client = file_system_client.get_directory_client(
            self.connection_settings.directory)

        if not directory_client.exists():
            file_system_client.create_directory(self.connection_settings.directory)
            directory_client = file_system_client.get_directory_client(
                self.connection_settings.directory)

        file_client = directory_client.create_file(self.connection_settings.file_name)
        file_contents = input_stream.read()
        file_client.upload_data(data=file_contents, overwrite=True)




""" Unit tests """

import json
import os
import unittest
from requests.auth import HTTPBasicAuth
from xcputils.ingestion.file import FileIngestor
from xcputils.ingestion.http import HttpIngestor, HttpMethod, HttpRequest
from xcputils.streaming.aws import AwsS3StreamReader, AwsS3ConnectionSettings
from xcputils.streaming.az import AdfsStreamReader, AdfsConnectionSettings

class TestHttpIngestor(unittest.TestCase):
    """ Test xcputils.ingest.http """

    def setUp(self):
        self.aws_connection_settings = AwsS3ConnectionSettings(
            bucket=os.environ['AWS_S3_BUCKET'],
            file_path="tests3streamconnector/folder/test.txt"
            )

        self.adfs_connection_settings = AdfsConnectionSettings(
            container="testadfsstreamconnector",
            file_name="test.txt",
            directory="folder"
            )

        self.file_path = "./test.txt"


    def tearDown(self):
        os.remove(self.file_path)


    def test_get(self):
        """ Test xcputils.ingest.http.get """

        http_ingestor = HttpIngestor(http_request=HttpRequest(url="https://postman-echo.com/ip"))

        result = http_ingestor.write_to_string()
        self.assertTrue("ip" in result, f"key 'ip' not in: {result}")

        http_ingestor.write_to_file(self.file_path)
        result = FileIngestor(self.file_path).write_to_string()
        self.assertTrue("ip" in result, f"key 'ip' not in: {result}")


        http_ingestor.write_to_aws_s3(
            bucket=self.aws_connection_settings.bucket,
            file_path=self.aws_connection_settings.file_path
            )
        result = AwsS3StreamReader(self.aws_connection_settings).read_str()
        self.assertTrue("ip" in result, f"key 'ip' not in: {result}")

        http_ingestor.write_to_adfs(
            container=self.adfs_connection_settings.container,
            file_name=self.adfs_connection_settings.file_name,
            directory=self.adfs_connection_settings.directory
            )
        result = AdfsStreamReader(self.adfs_connection_settings).read_str()
        self.assertTrue("ip" in result, f"key 'ip' not in: {result}")


    def test_get_html(self):
        """ Test xcputils.ingest.http.get HTML """

        http_ingestor = HttpIngestor(http_request=HttpRequest(url="https://postman-echo.com"))

        result = http_ingestor.write_to_string()
        self.assertEqual(result[0:15], "<!DOCTYPE html>")

        http_ingestor.write_to_file(self.file_path)
        result = FileIngestor(self.file_path).write_to_string()
        self.assertEqual(result[0:15], "<!DOCTYPE html>")


        http_ingestor.write_to_aws_s3(
            bucket=self.aws_connection_settings.bucket,
            file_path=self.aws_connection_settings.file_path
            )
        result = AwsS3StreamReader(self.aws_connection_settings).read_str()
        self.assertEqual(result[0:15], "<!DOCTYPE html>")

        http_ingestor.write_to_adfs(
            container=self.adfs_connection_settings.container,
            file_name=self.adfs_connection_settings.file_name,
            directory=self.adfs_connection_settings.directory
            )
        result = AdfsStreamReader(self.adfs_connection_settings).read_str()
        self.assertEqual(result[0:15], "<!DOCTYPE html>")



    def test_auth(self):
        """ Test xcputils.ingest.http.get authentication """

        http_ingestor = HttpIngestor(
            http_request=HttpRequest(
                url="https://postman-echo.com/basic-auth",
                auth=HTTPBasicAuth('postman', 'password')
                ))

        result = http_ingestor.write_to_string()
        result = json.loads(result)
        self.assertEqual(result["authenticated"], True, result)

        http_ingestor.write_to_file(self.file_path)
        result = FileIngestor(self.file_path).write_to_string()
        result = json.loads(result)
        self.assertEqual(result["authenticated"], True, result)


        http_ingestor.write_to_aws_s3(
            bucket=self.aws_connection_settings.bucket,
            file_path=self.aws_connection_settings.file_path
            )
        result = AwsS3StreamReader(self.aws_connection_settings).read_str()
        result = json.loads(result)
        self.assertEqual(result["authenticated"], True, result)

        http_ingestor.write_to_adfs(
            container=self.adfs_connection_settings.container,
            file_name=self.adfs_connection_settings.file_name,
            directory=self.adfs_connection_settings.directory
            )
        result = AdfsStreamReader(self.adfs_connection_settings).read_str()
        result = json.loads(result)
        self.assertEqual(result["authenticated"], True, result)


    def test_post(self):
        """ Test xcputils.ingest.http.post """

        data = {"test_key": "test_value"}

        http_ingestor = HttpIngestor(
            http_request=HttpRequest(
                url="https://postman-echo.com/post",
                method=HttpMethod.POST,
                body=data,
                ))

        result = http_ingestor.write_to_string()
        result = json.loads(result)
        self.assertEqual(result["data"], data)

        http_ingestor.write_to_file(self.file_path)
        result = FileIngestor(self.file_path).write_to_string()
        result = json.loads(result)
        self.assertEqual(result["data"], data)


        http_ingestor.write_to_aws_s3(
            bucket=self.aws_connection_settings.bucket,
            file_path=self.aws_connection_settings.file_path
            )
        result = AwsS3StreamReader(self.aws_connection_settings).read_str()
        result = json.loads(result)
        self.assertEqual(result["data"], data)

        http_ingestor.write_to_adfs(
            container=self.adfs_connection_settings.container,
            file_name=self.adfs_connection_settings.file_name,
            directory=self.adfs_connection_settings.directory
            )
        result = AdfsStreamReader(self.adfs_connection_settings).read_str()
        result = json.loads(result)
        self.assertEqual(result["data"], data)


if __name__ == '__main__':
    unittest.main()



""" Unit tests """

from io import BytesIO
import unittest
import datetime
from xcputils.ingestion.string import StringIngestor

from xcputils.streaming.az import AdfsStreamWriter, AdfsConnectionSettings, AdfsStreamReader


class TestAdfsStreamWriter(unittest.TestCase):
    """ Test  """


    def test_write_read(self):
        """ Test AdfsStreamConnector.read """

        connection_settings = AdfsConnectionSettings(
            container="testadfsstreamconnector",
            directory="folder",
            file_name="test.txt")

        writer = AdfsStreamWriter(connection_settings)

        payload = f"Testing.\n123.\næøåÆØÅ\n{datetime.datetime.now()}"

        with BytesIO() as stream:
            stream.write(payload.encode('utf-8'))
            stream.seek(0)
            writer.write(stream)

        reader = AdfsStreamReader(connection_settings)
        actual_payload = reader.read_str()

        self.assertEqual(actual_payload, payload)


if __name__ == '__main__':
    unittest.main()



""" Unit tests """

import unittest
from xcputils.batch import Batch, async_wrap


class BatchTestCase(unittest.TestCase):
    """ Test Batch """

    def test_batch(self):
        """ Test Batch append and callable """
        inc = lambda x: x + 1

        @async_wrap
        def inc_async(x):
            return x+1

        batch = Batch() \
            .append(inc, x=1) \
            .append_async(inc_async(x=2)) \
            .append_async(async_wrap(inc)(x=3))

        result = batch()
        self.assertEqual(result, [2, 3, 4])


if __name__ == '__main__':
    unittest.main()



""" HTTP Request callable """

from io import BytesIO
import json
import os
from typing import ContextManager
from enum import Enum
import copy
import requests
from xcputils.ingestion import Ingestor

from xcputils.streaming import StreamWriter


class HttpMethod(Enum):
    """ HTTP method """

    GET = "GET"
    POST = "POST"


class HttpRequest():
    """ HTTP request parameters """

    def __init__(
        self,
        url: str,
        method: HttpMethod = HttpMethod.GET,
        params: dict = None,
        body: dict = None,
        headers: dict = None,
        auth = None):

        self.url = url
        self.method = method
        self.params = {} if not params else params
        self. body = {} if not body else body
        self.headers = {} if not headers else headers
        self.auth = auth


class PaginationHandler:
    """ Paginated HTTP request pagination_handler """

    def __init__(
        self,
        page_size: int,
        page_size_param: str = "limit",
        data_property: str = "data",
        max_pages: int = 1000):

        self.page_size = page_size
        self.page_size_param = page_size_param
        self.data_property = data_property
        self.max_pages = max_pages


    def get_page_request(self, request: HttpRequest, page_number: int) -> HttpRequest:
        """ Compose HTTP request for a page """

        page_request = copy.deepcopy(request)
        page_request.params[self.page_size_param] = self.page_size
        page_request.params["offset"] = (page_number - 1) * self.page_size
        return page_request


    def is_last_page(self, payload: dict, page_number: int) -> bool:
        """ Check if last page """

        if page_number >= self.max_pages:
            return True

        if not self.data_property in payload.keys():
            return True

        if not isinstance(payload[self.data_property], list):
            return True

        if len(payload[self.data_property]) < self.page_size:
            return True

        return False


class HttpIngestor(Ingestor):
    """ HTTP ingestor """

    def __init__(
        self,
        http_request: HttpRequest = None,
        pagination_handler: PaginationHandler = None,
        stream_writer: StreamWriter = None,
        ):
        """ Constructor """

        super().__init__(stream_writer)

        self.http_request = http_request
        self.pagination_handler = pagination_handler

        self.http_methods = {
            HttpMethod.GET: self._get,
            HttpMethod.POST: self._post
            }


    def with_pagination(
        self,
        page_size: int=1000,
        page_size_param: str = "limit",
        data_property: str = "data",
        max_pages: int = 1000) -> "HttpIngestor":
        """ Pagination parameters"""

        self.pagination_handler = PaginationHandler(
            page_size=page_size,
            page_size_param=page_size_param,
            data_property=data_property,
            max_pages=max_pages)

        return self


    def with_stream_writer(
        self,
        stream_writer: StreamWriter) -> "HttpIngestor":
        """ Set stream writer """

        self.stream_writer = stream_writer

        return self


    def _get(self, session, request, stream):
        return session.get(
            url=request.url,
            params=request.params,
            headers=request.headers,
            auth=request.auth,
            stream=stream)


    def _post(self, session, request, stream):
        return session.post(
            url=request.url,
            params=request.params,
            data=json.dumps(request.body),
            headers=request.headers,
            auth=request.auth,
            stream=stream)


    def ingest(self):
        """ Ingest """

        if self.pagination_handler:
            self.ingest_paginated()
            return

        with requests.Session() as session:
            http_method = self.http_methods[self.http_request.method]
            response = http_method(session=session, request=self.http_request, stream=True)

            response.raise_for_status()

            if isinstance(response, ContextManager):
                with response as part:
                    part.raw.decode_content = True
                    self.stream_writer.write(part.raw)
            else:
                response.raw.decode_content = True
                self.stream_writer.write(response.raw)


    def ingest_paginated(self):
        """ Ingest paginated """

        page_number = 1
        is_last_page = False

        while not is_last_page:

            page_request = self.pagination_handler \
                .get_page_request(self.http_request, page_number=page_number)

            with requests.Session() as session:
                http_method = self.http_methods[self.http_request.method]
                response = http_method(session=session, request=page_request, stream=False)

                response.raise_for_status()
                payload = response.json()

                hold_filename = self.stream_writer.get_file_path()
                filename, ext = os.path.splitext(hold_filename)
                self.stream_writer.set_file_path(f"{filename}.{page_number}{ext}")

                with BytesIO() as stream:
                    stream.write(json.dumps(payload).encode('utf-8'))
                    stream.seek(0)
                    self.stream_writer.write(stream)

                self.stream_writer.set_file_path(hold_filename)

            is_last_page = self.pagination_handler \
                .is_last_page(payload=payload, page_number=page_number)

            page_number += 1



""" AWS S3 file """

import os
from typing import Any
import boto3
from boto3.s3.transfer import TransferConfig

from xcputils.streaming import StreamReader, StreamWriter


class AwsS3ConnectionSettings():
    """ AWS S3 connection settings """

    def __init__(self,
                 bucket: str,
                 file_path: str,
                 aws_access_key_id: str = None,
                 aws_secret_access_key: str = None,
                 aws_session_token: str = None,
                 aws_region_name: str = None,
                 ):

        self.bucket = bucket
        self.file_path = file_path
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.aws_region_name = aws_region_name


    def get_client(self):
        """ Get S3 client """

        return boto3.Session(
            aws_access_key_id=self.aws_access_key_id
                if self.aws_access_key_id
                else os.getenv("AWS_ACCESS_KEY_ID", None),
            aws_secret_access_key=self.aws_secret_access_key
                if self.aws_secret_access_key
                else os.getenv("AWS_SECRET_ACCESS_KEY", None),
            aws_session_token=self.aws_session_token
                if self.aws_session_token
                else os.getenv("AWS_SESSION_TOKEN", None),
            region_name=self.aws_region_name
                if self.aws_region_name
                else os.getenv("AWS_DEFAULT_REGION", None)
            ).client('s3')


class AwsS3StreamReader(StreamReader):
    """ AWS S3 stream writer """

    def __init__(self, connection_settings: AwsS3ConnectionSettings):
        super().__init__()
        self.connection_settings = connection_settings


    def read(self, output_stream: Any):
        """ Read from stream """

        client = self.connection_settings.get_client()

        client.download_fileobj(
            self.connection_settings.bucket,
            self.connection_settings.file_path,
            output_stream)


class AwsS3StreamWriter(StreamWriter):
    """ AWS S3 stream writer """

    def __init__(self, connection_settings: AwsS3ConnectionSettings):
        super().__init__()
        self.connection_settings = connection_settings


    def get_file_path(self) -> str:
        """ Get filename """

        return self.connection_settings.file_path


    def set_file_path(self, file_path: str):
        """ Set filename """

        self.connection_settings.file_path = file_path


    def write(self, input_stream: Any):
        """ Write to stream """

        conf = TransferConfig(multipart_threshold=10000, max_concurrency=4)

        client = self.connection_settings.get_client()

        client.upload_fileobj(
            input_stream,
            self.connection_settings.bucket,
            self.connection_settings.file_path,
            Config=conf)



import io
import json
from unittest import mock


def mock_response(
        status=200,
        content="CONTENT",
        json_data=None,
        raise_for_status=None):
    """ Compose mock response """

    mock_resp = mock.Mock()

    mock_resp.raise_for_status = mock.Mock()
    if raise_for_status:
        mock_resp.raise_for_status.side_effect = raise_for_status

    mock_resp.status_code = status
    mock_resp.content = content

    if json_data:
        mock_resp.json.return_value = json_data
        mock_resp.content = json.dumps(json_data)

    mock_resp.raw = io.BytesIO(mock_resp.content.encode("utf-8"))

    return mock_resp



""" Test File Ingestion """

import os
import unittest

from xcputils.ingestion.file import FileIngestor


class TestFileIngestor(unittest.TestCase):
    """ Test FileIngestor """

    def setUp(self):
        self.file_content = "Hello, world! This is a test file."
        self.file_path = "./test_file.txt"
        with open(self.file_path, "w", encoding="utf-8") as file:
            file.write(self.file_content)


    def tearDown(self):
        os.remove(self.file_path)


    def test_ingest(self):
        """ Test ingest """

        result = FileIngestor(self.file_path).write_to_string()
        self.assertEqual(result, self.file_content)


    def test_init(self):
        """ Test init """

        ingestor = FileIngestor(self.file_path)
        self.assertEqual(ingestor.file_path, self.file_path)
        self.assertEqual(ingestor.stream_writer, None)



""" String ingestion """

from io import BytesIO
from xcputils.ingestion import Ingestor
from xcputils.streaming import StreamWriter


class StringIngestor(Ingestor):
    """ Ingest from string """

    def __init__(
        self,
        data: str,
        stream_writer: StreamWriter = None,
        ):

        super().__init__(stream_writer)

        self.data = data


    def ingest(self):
        """ Ingest """

        with BytesIO() as stream:
            stream.write(self.data.encode('utf-8'))
            stream.seek(0)
            self.stream_writer.write(stream)



""" Test File Ingestion """

import datetime
from io import BytesIO
import os
import unittest
from xcputils.ingestion.file import FileIngestor

from xcputils.streaming.file import FileStreamWriter



class TestStringStreamWriter(unittest.TestCase):
    """ Test FileStreamWriter """

    def setUp(self):
        self.file_path = "./test_file.txt"


    def tearDown(self):
        os.remove(self.file_path)


    def test_write(self):
        """ Test write """

        payload = f"Testing.\n123.\næøåÆØÅ\n{datetime.datetime.now()}"
        writer = FileStreamWriter(self.file_path)

        with BytesIO() as stream:
            stream.write(payload.encode('utf-8'))
            stream.seek(0)
            writer.write(stream)

        result = FileIngestor(self.file_path).write_to_string()

        self.assertEqual(result, payload)



""" Unit tests """

import json
import unittest
from unittest.mock import patch
from test.unit import mock_response
from requests import Session
from xcputils.ingestion.http import HttpIngestor, HttpRequest
from xcputils.streaming.string import StringStreamWriter


class TestPaginatedHttpIngestor(unittest.TestCase):
    """ Test xcputils.ingest.http """


    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)
        self.stream_writer = StringStreamWriter()



    @patch.object(Session, "get")
    def test_get(self, mock_get):
        """ Test HttpIngestor with pagination """

        mock_get.side_effect = [
            mock_response(json_data={"data": [1, 2, 3]}),
            mock_response(json_data={"data": [4, 5, 6]}),
            mock_response(json_data={"data": [7, 8]}),
        ]

        result = HttpIngestor(
            http_request=HttpRequest(url="https://mock.com/ip")) \
            .with_pagination(page_size=3) \
            .write_to_string()

        result = [json.loads(line) for line in result.splitlines()]
        print(result)
        self.assertTrue(result == [
            {"data": [1, 2, 3]},
            {"data": [4, 5, 6]},
            {"data": [7, 8]},
            ])


if __name__ == "__main__":
    unittest.main()



""" Checkpoints """

class Checkpoint:
    """ Checkpoint base class """

    def get(self, default: str = None) -> str:
        """ Get checkpoint """

    def set(self, value: str):
        """ Set checkpoint """

    def reset(self):
        """ Reset checkpoint """
        self.set("")        



""" Package xcputils """

from xcputils.checkpoints import Checkpoint
from xcputils.checkpoints.aws import AwsS3Checkpoint
from xcputils.checkpoints.az import AdfsCheckpoint
from xcputils.checkpoints.file import FileCheckpoint
from xcputils.ingestion.aws import AwsS3Ingestor
from xcputils.ingestion.az import AdfsIngestor
from xcputils.ingestion.file import FileIngestor
from xcputils.ingestion.ftp import FtpIngestor
from xcputils.ingestion.http import HttpIngestor, HttpMethod, HttpRequest
from xcputils.ingestion.string import StringIngestor
from xcputils.streaming.aws import AwsS3ConnectionSettings
from xcputils.streaming.az import AdfsConnectionSettings


class XCPUtils():
    """ Utilities for copying data """


    def read_from_string(self, data: str) -> StringIngestor:
        """ Ingest from string """

        return StringIngestor(data)


    def read_from_file(self, file_path: str) -> FileIngestor:
        """ Ingest from file """

        return FileIngestor(file_path=file_path)


    def read_from_http(
        self,
        url: str,
        method: HttpMethod = HttpMethod.GET,
        params: dict = None,
        body: dict = None,
        headers: dict = None,
        auth = None,
        ) -> HttpIngestor:
        """ Ingest from HTTP """

        return HttpIngestor(
            http_request=HttpRequest(
                url=url,
                method=method,
                params=params,
                body=body,
                headers=headers,
                auth=auth)
        )


    def read_from_aws_s3(
        self,
        bucket: str,
        file_path: str,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
        aws_session_token: str = None,
        aws_region_name: str = None,
    ) -> AwsS3Ingestor:
        """ Ingest from AWS S3 """

        return AwsS3Ingestor(
            connection_settings=AwsS3ConnectionSettings(
                bucket=bucket,
                file_path=file_path,
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
                aws_region_name=aws_region_name,
            )
        )


    def read_from_adfs(
        self,
        container: str,
        file_name: str,
        directory: str,
        storage_account_name: str = None,
        tenant_id: str = None,
        client_id: str = None,
        client_secret: str = None,
    ) -> AdfsIngestor:
        """ Ingest from AWS S3 """

        return AdfsIngestor(
            connection_settings=AdfsConnectionSettings(
                container=container,
                file_name=file_name,
                directory=directory,
                storage_account_name=storage_account_name,
                tenant_id=tenant_id,
                client_id=client_id,
                client_secret=client_secret
            )
        )


    def read_from_ftp(
        self,
        url: str,
        folder: str = None,
        file_name: str = None,
        user: str = None,
        password: str = None,
    ) -> FtpIngestor:
        """ Ingest frm FTP """

        return FtpIngestor(
            url=url,
            folder=folder,
            file_name=file_name,
            user=user,
            password=password,
            )

    def get_checkpoint_file(self, name: str, directory: str) -> Checkpoint:
        """ Get file checkpoint """

        return FileCheckpoint(name=name,directory=directory)


    def get_checkpoint_adfs(
        self, 
        name: str,
        container: str,
        directory: str,
        storage_account_name: str = None,
        tenant_id: str = None,
        client_id: str = None,
        client_secret: str = None,
        ) -> Checkpoint:
        """ Get ADFS checkpoint """

        return AdfsCheckpoint(
            name=name,
            container=container,
            directory=directory,
            storage_account_name=storage_account_name,
            tenant_id=tenant_id,
            client_id=client_id,
            client_secret=client_secret,
        )


    def get_checkpoint_awss3(
        self, 
        name: str,
        bucket: str,
        directory: str,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
        aws_session_token: str = None,
        aws_region_name: str = None,
        ):
        """ Get AWS S3 checkpoint """

        return AwsS3Checkpoint(
            name=name,
            bucket=bucket,
            directory=directory,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            aws_region_name=aws_region_name,
        )



""" Connectors to read and write streams """

from io import BytesIO
import tempfile
from typing import Any


class StreamReader():
    """ Stream reader base class """


    def read(self, output_stream: Any):
        """ Read from stream """


    def read_str(self) -> str:
        """ Read from stream to a string """

        with tempfile.TemporaryFile() as data:
            self.read(data)
            data.seek(0)
            return data.read().decode('utf-8')


class StreamWriter():
    """ Stream writer base class """


    def get_file_path(self) -> str:
        """ Get filename """
        return ""


    def set_file_path(self, file_path: str):
        """ Set filename """


    def write(self, input_stream: Any):
        """ Write stream """



""" Test File Ingestion """

import datetime
import unittest
from xcputils.ingestion.string import StringIngestor


class TestStringIngestor(unittest.TestCase):
    """ Test StringIngestor """

    def test_ingest(self):
        """ Test ingest """

        payload = f"Testing.\n123.\næøåÆØÅ\n{datetime.datetime.now()}"
        result = StringIngestor(payload).write_to_string()

        self.assertEqual(result, payload)



""" Unit tests """

import datetime
import os
import unittest
from xcputils.ingestion.aws import AwsS3Ingestor
from xcputils.ingestion.string import StringIngestor
from xcputils.streaming.aws import AwsS3ConnectionSettings


class TestAwsS3Ingestor(unittest.TestCase):
    """ Test xcputils.ingestion.aws.AwsS3Ingestor """


    def test_ingest(self):
        """ xcputils.streaming.aws.AwsS3Ingestor """

        bucket = os.environ['AWS_S3_BUCKET']
        file_path = "tests3ingestor/test.txt"
        payload = f"Testing.\n123.\næøåÆØÅ\n{datetime.datetime.now()}"


        StringIngestor(data=payload) \
            .write_to_aws_s3(bucket, file_path)

        result = AwsS3Ingestor(AwsS3ConnectionSettings(bucket, file_path)).write_to_string()

        self.assertEqual(result, payload)


if __name__ == '__main__':
    unittest.main()



""" Ingest from AWS S3 """

from io import BytesIO
from xcputils.ingestion import Ingestor
from xcputils.streaming import StreamWriter
from xcputils.streaming.aws import AwsS3ConnectionSettings, AwsS3StreamReader


class AwsS3Ingestor(Ingestor):
    """ Ingest from AWS S3 """

    def __init__(
        self,
        connection_settings: AwsS3ConnectionSettings,
        stream_writer: StreamWriter = None,
        ):

        super().__init__(stream_writer)
        self.connection_settings = connection_settings


    def ingest(self):
        """ Ingest """

        with BytesIO() as stream:
            AwsS3StreamReader(self.connection_settings) \
                .read(stream)
            stream.seek(0)
            self.stream_writer.write(stream)



""" Unit tests """

import os
import unittest
from xcputils.ingestion.ftp import FtpIngestor
from xcputils.streaming.aws import AwsS3ConnectionSettings, AwsS3StreamReader
from xcputils.streaming.az import AdfsConnectionSettings, AdfsStreamReader


class TestFtpIngestor(unittest.TestCase):
    """ Test xcputils.ingestion.ftp.FtpIngestor """

    def test_ingest(self):
        """ xcputils.streaming.ftp.FtpIngestor """

        download_path = "./data/dlptest"
        os.makedirs(name=download_path, exist_ok=True)

        ingestor = FtpIngestor(
            url="test.rebex.net",
            folder=None,
            file_name=None,
            user="demo",
            password="password",
            )

        ingestor.write_to_file(download_path)

        files = os.listdir(download_path)
        file_name = files[0]
        os.system(f"ls -l {download_path}")
        self.assertTrue(len(files) > 0)
        os.system(f"rm -r {download_path}")
        os.makedirs(name=download_path, exist_ok=True)

        ingestor.file_name = file_name
        ingestor.write_to_file(download_path)
        files = os.listdir(download_path)
        os.system(f"ls -l {download_path}")
        self.assertTrue(len(files) == 1)
        os.system(f"rm -r {download_path}")

        result = ingestor.write_to_string()
        self.assertTrue(len(result) > 0)

        ingestor.write_to_aws_s3(
            bucket=os.environ['AWS_S3_BUCKET'],
            file_path="testfptingetor/temp"
            )
        result = AwsS3StreamReader(
            connection_settings=AwsS3ConnectionSettings(
                bucket=os.environ['AWS_S3_BUCKET'],
                file_path=f"testfptingetor/temp/{file_name}"
            )
        ).read_str()
        self.assertTrue(len(result) > 0)

        ingestor.write_to_adfs(
            container="testfptingetor",
            file_name="",
            directory="temp"
            )
        result = AdfsStreamReader(
            connection_settings=AdfsConnectionSettings(
                container="testfptingetor",
                file_name=file_name,
                directory="temp"
            )
        ).read_str()
        self.assertTrue(len(result) > 0)


if __name__ == '__main__':
    unittest.main()
